# Adapted from: https://github.com/amjith/fuzzyfinder
# https://blog.amjith.com/fuzzyfinder-in-10-lines-of-python
# It sorts the results differently from the original, with matches
# ordered by compactness and then index in original list

import re


def fuzzyfinder(input, collection):
    """
    Args:
        input (str): A partial string which is typically entered by a user.
        collection (iterable): A collection of strings which will be filtered
                               based on the `input`.

    Returns:
        suggestions (generator): A generator object that produces a list of
            suggestions narrowed down from `collection` using the `input`.
    """
    suggestions = []
    input = str(input) if not isinstance(input, str) else input
    pat = ".*?".join(map(re.escape, input))
    pat = "(?=({0}))".format(
        pat
    )  # lookahead regex to manage overlapping matches
    regex = re.compile(pat, re.IGNORECASE)
    for idx, item in enumerate(collection):
        r = list(regex.finditer(item))
        if r:
            best = min(r, key=lambda x: len(x.group(1)))  # find shortest match
            suggestions.append((len(best.group(1)), idx, item))

    # print(regex)
    # return (z for z in sorted(suggestions))
    return (z[-1] for z in sorted(suggestions))



import asyncio
import builtins
import ctypes
import inspect
import json
import os
import random
import re
import threading
import traceback
from collections import defaultdict, deque
from contextlib import contextmanager
from contextvars import ContextVar
from itertools import count

from hrepr import H, Tag, hrepr

from .config import mayread, maywrite
from .fzf import fuzzyfinder
from .registry import callback_registry

_c = count(1)


class NoPatternException(Exception):
    pass


class CommandDispatcher:
    def __init__(self, patterns):
        self.patterns = [
            (re.compile(pattern, re.MULTILINE | re.DOTALL), fn)
            for pattern, fn in patterns.items()
        ]

    def __call__(self, expr):
        for (pattern, fn) in self.patterns:
            if m := pattern.fullmatch(expr):
                try:
                    return fn(m[0], *m.groups())
                except NotImplementedError:
                    continue
        else:
            raise NoPatternException(f"'{expr}' matches no known pattern")


class ThreadKilledException(Exception):
    pass


class SnektalkInterrupt(Exception):
    pass


def kill_thread(thread, exctype=ThreadKilledException):
    ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(thread.ident), ctypes.py_object(exctype)
    )


class KillableThread(threading.Thread):
    @property
    def dead(self):
        return not self.is_alive() or getattr(self, "_dead", False)

    kill = kill_thread


class NamedThreads:
    @classmethod
    def current(cls):
        thread = threading.current_thread()
        if isinstance(thread, KillableThread):
            return thread
        else:
            return None

    def __init__(self):
        self.words = [
            word
            for word in open(
                os.path.join(os.path.dirname(__file__), "words.txt")
            )
            .read()
            .split("\n")
            if word
        ]
        random.shuffle(self.words)
        self.threads = {}
        self.count = count(1)

    def run_in_thread(self, fn, session):
        def run():
            reason = " finished"
            self.threads[word] = thread
            with session.set_context():
                with new_evalid():
                    session.queue_result(
                        H.div("Starting thread ", H.strong(word)), type="info",
                    )
                    try:
                        fn()
                    except ThreadKilledException:
                        reason = " killed"
                    except Exception as e:
                        session.queue_result(e, type="exception")
                    finally:
                        thread._dead = True
                        session._clean_owners()
                        del self.threads[word]
                        self.words.append(word)
                        session.queue_result(
                            H.div("Thread ", H.strong(word), reason),
                            type="info",
                        )

        thread = KillableThread(target=run, daemon=True)

        if not self.words:
            self.words.append(f"t{next(self.count)}")
        word = self.words.pop()

        thread.start()
        return word, thread


threads = NamedThreads()
_current_session = ContextVar("current_session", default=None)
_current_print_session = ContextVar("current_print_session", default=None)
_current_evalid = ContextVar("current_evalid", default=None)


def current_session():
    return _current_session.get()


def current_print_session():
    return _current_print_session.get()


@contextmanager
def new_evalid():
    token = _current_evalid.set(next(_c))
    try:
        yield
    finally:
        _current_evalid.reset(token)


class Lib:
    def __init__(self, session):
        self._export = None
        self.session = session

    def method_stop(self):
        thread = self.session.owner or threading.main_thread()
        kill_thread(thread, SnektalkInterrupt)

    def method_history_navigate(self, delta, query):
        return self.session.history.navigate(delta, query)

    def method_populate_popup(self, name, query):
        if name == "history":
            return [
                {"text": entry} for entry in self.session.history.search(query)
            ]
        else:
            return [{"text": "main"}] + [
                {"text": name} for name, thread in threads.threads.items()
            ]

    def export(self):
        if self._export:
            return self._export

        self._export = rval = {}
        for method_name in dir(type(self)):
            if method_name.startswith("method_"):
                method = getattr(self, method_name)
                method_id = callback_registry.register(method)
                rval[method_name[7:]] = method_id

        return rval


class History:
    def __init__(self, history_file):
        self.history_file = history_file
        self.read()

    def read(self):
        self.history = deque(mayread(self.history_file, default=[]))
        self.query_results = self.history
        self.current_query = ""
        self.reset_cursor()

    def save(self):
        maywrite(self.history_file, list(self.history)[:1000])

    def _clamp(self, x):
        return max(-1, min(len(self.history) - 1, x))

    def append(self, entry):
        if not self.history or self.history[0] != entry:
            self.history.appendleft(entry)
        self.reset_cursor()

    def reset_cursor(self):
        self.cursor = -1

    def navigate(self, delta, query):
        if self.cursor == -1 or query != self.query_results[self.cursor]:
            if query == "":
                self.query_results = self.history
                self.current_query = ""
            else:
                self.query_results = list(self.search(query))
                self.current_query = query

        prev = self.cursor
        self.cursor = max(
            -1, min(len(self.query_results) - 1, self.cursor - delta)
        )
        if self.cursor == -1:
            return self.current_query
        elif prev != self.cursor:
            return self.query_results[self.cursor]
        else:
            return None

    def search(self, query):
        return fuzzyfinder(query, self.history)


class Session:
    def __init__(self, socket=None, history_file=None):
        self.tempkeep = deque(maxlen=10)
        self.lib = Lib(self)
        self.blt = vars(builtins)
        self.idmap = {}
        self.varcount = count(1)
        self.socket = socket
        self.sent_resources = set()
        self.last_prompt = ""
        self.last_nav = ""
        self.in_queue = deque()
        self.out_queue = deque()
        self.semaphores = defaultdict(lambda: threading.Semaphore(value=0))
        self.navs = {}
        self.owners = []
        self.history = History(history_file)
        self.dispatch = CommandDispatcher(
            {
                "/attach[ \n]?(.*)": self.submit_command_attach,
                "/detach[ \n]?(.*)": self.submit_command_detach,
                "/kill[ \n]?(.*)": self.submit_command_kill,
            }
        )
        self._token = None
        self._tokenp = None

    #############
    # Utilities #
    #############

    @property
    def owner(self):
        return self.owners[-1] if self.owners else None

    def _clean_owners(self):
        while self.owners and (curr := self.owners[-1]) and curr.dead:
            del self.navs[curr]
            self.pop_owner()

    def _acquire_all(self):
        current = self.semaphores[self.owner]
        while current.acquire(blocking=False):
            pass

    def _release_all(self):
        current = self.semaphores[self.owner]
        if self.in_queue:
            current.release(len(self.in_queue))

    def _current_prompt(self):
        for act in self.navs.get(self.owner, {}).values():
            act()

    def push_owner(self, thread):
        self._acquire_all()
        self.owners.append(thread)
        self._release_all()
        self._current_prompt()

    def pop_owner(self):
        self._acquire_all()
        if self.owners:
            self.owners.pop()
        self._release_all()
        self._current_prompt()

    def enter(self, capture_print=True):
        self._token = _current_session.set(self)
        if capture_print:
            self._tokenp = _current_print_session.set(self)

    def exit(self):
        _current_session.reset(self._token)
        if self._tokenp:
            _current_print_session.reset(self._tokenp)

    @contextmanager
    def set_context(self):
        token = _current_session.set(self)
        tokenp = _current_print_session.set(self)
        try:
            yield
        finally:
            _current_session.reset(token)
            _current_print_session.reset(tokenp)

    def set_globals(self, glb):
        self.glb = glb

    def newvar(self):
        """Create a new variable."""
        return f"_{next(self.varcount)}"

    def getvar(self, obj):
        """Get the variable name corresponding to the object.

        If the object is not already associated to a variable, one
        will be created and set in the global scope.
        """
        ido = id(obj)
        if ido in self.idmap:
            varname = self.idmap[ido]
        else:
            varname = self.newvar()
            self.idmap[ido] = varname
        self.blt[varname] = obj
        return varname

    def represent(self, typ, result):
        if isinstance(result, Tag):
            return typ, result

        try:
            html = hrepr(result)
        except Exception as exc:
            try:
                html = hrepr(exc)
            except Exception:
                html = H.pre(
                    traceback.format_exception(
                        builtins.type(exc), exc, exc.__traceback__
                    )
                )
                typ = "hrepr_exception"
        return typ, html

    ################
    # Proceed/next #
    ################

    def set_prompt(self, prompt):
        self.add_nav_action(
            "prompt", lambda: self.queue(command="set_mode", html=prompt)
        )

    def set_nav(self, nav):
        self.add_nav_action(
            "nav",
            lambda: self.queue(
                command="set_nav",
                value=hrepr(nav, interactive=False),
                navid=id(nav),
            ),
        )

    def add_nav_action(self, name, action):
        self.navs.setdefault(NamedThreads.current(), {})[name] = action

    @contextmanager
    def prompt(self, prompt="", nav=H.span()):
        self.set_prompt(prompt)
        self.set_nav(nav)
        self._clean_owners()
        self._current_prompt()
        expr = self.next()
        with self.set_context():
            with new_evalid():
                yield expr

    def next(self):
        self.semaphores[NamedThreads.current()].acquire()
        return self.in_queue.popleft()

    ###########
    # Sending #
    ###########

    def bind(self, socket):
        self.loop = asyncio.get_running_loop()
        self.socket = socket
        self.sent_resources = set()
        while self.out_queue:
            self.schedule(self.send(**self.out_queue.popleft()))
        self.queue(command="set_lib", lib=self.lib.export())
        self.submit({"command": "noop"})

    async def send(self, process=True, **command):
        """Send a command to the client, plus any resources.

        Any field that is a Tag and contains resources will send
        resource commands to the client to load these resources.
        A resource is only sent once, the first time it is needed.
        """
        if process:
            resources = []
            for k, v in command.items():
                if isinstance(v, Tag):
                    resources.extend(v.collect_resources())
                    command[k] = str(v)

            for resource in resources:
                if resource not in self.sent_resources:
                    await self.socket.send(
                        json.dumps(
                            {"command": "resource", "value": str(resource)}
                        )
                    )
                    self.sent_resources.add(resource)

            evalid = _current_evalid.get()
            if evalid is not None:
                command["evalid"] = evalid

        await self.socket.send(json.dumps(command))

    def schedule(self, fn):
        self.loop.call_soon_threadsafe(lambda: self.loop.create_task(fn))

    def queue(self, **command):
        """Queue a command to the client, plus any resources.

        This queues the command using the session's asyncio loop.
        """
        if self.socket is None or self.out_queue:
            self.out_queue.append(command)
        else:
            self.schedule(self.send(**command))

    def queue_result(self, result, *, type):
        type, html = self.represent(type, result)
        self.tempkeep.append(result)
        self.queue(command="result", value=html, type=type)

    ############
    # Commands #
    ############

    def submit_command_attach(self, expr, tname):
        tname = tname.strip()
        self.queue(command="echo", value=expr, process=False)
        if tname == "main":
            self.push_owner(None)
        elif tname in threads.threads:
            thread = threads.threads[tname]
            self.push_owner(thread)
        else:
            self.queue(
                command="result",
                value=f"No thread named {tname}"
                if tname
                else "Please provide the name of the thread to attach to",
                type="exception",
            )

    def submit_command_detach(self, expr, arg):
        self.queue(command="echo", value=expr, process=False)
        assert not arg.strip()
        self.pop_owner()

    def submit_command_kill(self, expr, tname):
        tname = tname.strip()
        self.queue(command="echo", value=expr, process=False)
        if tname in threads.threads:
            thread = threads.threads[tname]
            thread.kill()
            self.queue(
                command="result",
                value=f"Sent an exception to {tname}. It should terminate as soon as possible.",
                type="print",
            )
        else:
            self.queue(
                command="result",
                value=f"No thread named {tname}"
                if tname
                else "Please provide the name of the thread to kill",
                type="exception",
            )

    def submit(self, data):
        self.in_queue.append(data)
        self.semaphores[self.owner].release()

    async def recv(self, **command):
        cmd = command.pop("command", "none")
        meth = getattr(self, f"command_{cmd}", None)
        await meth(**command)

    async def command_submit(self, *, expr):
        if expr.strip():
            self.history.append(expr)
        try:
            self.dispatch(expr)
        except NoPatternException:
            self.submit({"command": "expr", "expr": expr})

    async def command_callback(self, *, id, response_id, arguments):
        try:
            cb = callback_registry.resolve(int(id))
        except KeyError:
            self.queue(
                command="status",
                type="error",
                value="value is unavailable; it might have been garbage-collected",
            )
            return

        try:
            with self.set_context():
                if inspect.isawaitable(cb) or inspect.iscoroutinefunction(cb):
                    result = await cb(*arguments)
                else:
                    result = cb(*arguments)

            await self.send(
                command="response", value=result, response_id=response_id
            )

        except Exception as exc:
            import traceback

            traceback.print_exc()
            await self.send(
                command="response",
                error={
                    "type": type(exc).__name__,
                    "message": str(exc.args[0]) if exc.args else None,
                },
                response_id=response_id,
            )

    def atexit(self):
        self.history.save()



import atexit
import errno
import json
import os
import random
import re
import socket
import subprocess
import threading
import webbrowser

import jurigged
from jurigged import codetools
from sanic import Sanic, response

from .config import get_config_path
from .network import create_inet, create_socket
from .repr import inject
from .session import Session

here = os.path.dirname(__file__)
assets_path = os.path.join(here, "assets")


def status_logger(sess):
    def log(event):
        if isinstance(event, codetools.UpdateOperation) and not isinstance(
            event.defn, codetools.FunctionDefinition
        ):
            return
        else:
            sess.queue(
                command="status", type="normal", value=str(event),
            )

    return log


def _launch(port=None, sock=None, open_browser=True, template={}, sess=None):
    if port is not None and sock is not None:
        raise ValueError("Cannot specify both port and socket")
    elif sock is not None:
        sock = create_socket(sock)
    elif port is None:
        sock = create_inet()
        host, port = sock.getsockname()
    else:
        host = "localhost"

    app = Sanic("snektalk")
    app.static("/favicon.ico", f"{assets_path}/favicon.ico")
    app.static("/lib/", f"{assets_path}/lib/")
    app.static("/scripts/", f"{assets_path}/scripts/")
    app.static("/style/", f"{assets_path}/style/")
    app.static("/fs/", "/")

    @app.route("/")
    async def index(request):
        index = open(os.path.join(assets_path, "index.html")).read()
        index = re.sub(
            r"{{([^{}]+)}}", lambda m: template.get(m[1], f"!!{m[1]}"), index
        )
        return response.html(index)

    @app.route("/status")
    async def status(request):
        return response.json({"status": "OK"})

    @app.websocket("/sktk")
    async def feed(request, ws):
        sess.bind(ws)
        while True:
            command = json.loads(await ws.recv())
            if sess.socket is ws:
                await sess.recv(**command)
            else:
                await ws.send(
                    json.dumps(
                        {
                            "command": "status",
                            "type": "error",
                            "value": "this connection was closed or pre-empted",
                        }
                    )
                )
                break

    if open_browser and port is not None:

        @app.listener("after_server_start")
        async def launch_browser(app, loop):
            webbrowser.open(f"http://{host}:{port}/")

    atexit.register(app.stop)
    atexit.register(sess.atexit)
    if port is not None:
        print(f"Start server at: http://{host}:{port}/")
    app.run(sock=sock, register_sys_signals=False)


def serve(watch_args=None, **kwargs):
    sess = Session(history_file=get_config_path("history.json"))
    if watch_args is not None:
        jurigged.watch(**watch_args, logger=status_logger(sess))

    def _start_server():
        _launch(sess=sess, **kwargs)

    thread = threading.Thread(target=_start_server, daemon=True)
    thread.start()

    sess.enter()
    inject()
    return sess



import ast
import functools
import re
import subprocess
import sys
import time
from types import ModuleType

from hrepr import H
from jurigged import CodeFile, registry
from jurigged.recode import virtual_file

from .feat.edit import edit
from .session import SnektalkInterrupt, current_session, new_evalid, threads
from .version import version

cmd_rx = re.compile(r"/([^ \n]+)([ \n].*)?", re.MULTILINE | re.DOTALL)


class StopEvaluator(Exception):
    pass


def safe_fail(fn):
    @functools.wraps(fn)
    def deco(self, *args, **kwargs):
        try:
            fn(self, *args, **kwargs)
        except Exception as e:
            self.session.blt["_"] = e
            self.session.blt["$$exc_info"] = sys.exc_info()
            self.session.queue_result(e, type="exception")

    return deco


def evaluate(expr, glb, lcl):
    filename = virtual_file("repl", expr)
    mname = glb.get("__name__", None)
    cf = CodeFile(filename=filename, source=expr, module_name=mname)
    registry.cache[filename] = cf

    tree = ast.parse(expr)
    assert isinstance(tree, ast.Module)
    assert len(tree.body) > 0
    *body, last = tree.body
    if isinstance(last, ast.Expr):
        last = ast.Expression(last.value)
    else:
        body.append(last)
        last = None
    for stmt in body:
        compiled = compile(
            ast.Module(body=[stmt], type_ignores=[]),
            mode="exec",
            filename=filename,
        )
        exec(compiled, glb, lcl)
    if last:
        compiled = compile(last, mode="eval", filename=filename)
        rval = eval(compiled, glb, lcl)
    else:
        rval = None

    cf.associate(glb)
    return rval


class Evaluator:
    def __init__(self, module, glb, lcl, session, prompt=">>>"):
        self.prompt = prompt
        self.session = session
        if module is None:
            module = ModuleType("__main__")
            module.__dict__.update(glb)
        self.module = module
        self.glb = glb
        self.lcl = lcl

    def eval(self, expr, glb=None, lcl=None):
        if glb is None:
            glb = self.glb
        if lcl is None:
            lcl = self.lcl if self.lcl is not None else glb
        return evaluate(expr, glb, lcl)

    def format_modname(self):
        modname = getattr(self.module, "__name__", "<module>")
        return H.span["snek-interpreter-in"](modname)

    shorthand = {
        "q": "quit",
        "d": "debug",
    }

    @safe_fail
    def command_eval(self, expr, glb=None, lcl=None):
        assert isinstance(expr, str)
        expr = expr.lstrip()

        self.session.queue(command="echo", value=expr, process=False)

        result = self.eval(expr, glb, lcl)
        typ = "statement" if result is None else "expression"

        self.session.blt["_"] = result

        self.session.queue_result(result, type=typ)

    @safe_fail
    def command_dir(self, expr, glb=None, lcl=None):
        from .repr import hdir

        expr = expr.lstrip()

        self.session.queue(command="echo", value=f"/dir {expr}", process=False)

        result = hdir(self.eval(expr, glb, lcl))
        typ = "statement" if result is None else "expression"

        self.session.blt["_"] = result

        self.session.queue_result(result, type=typ)

    @safe_fail
    def command_edit(self, expr, glb, lcl):
        expr = expr.lstrip()
        self.session.queue(command="echo", value=f"/edit {expr}", process=False)
        obj = self.eval(expr, glb, lcl)
        self.session.queue_result(
            edit(obj, evaluator=self, autofocus=True), type="expression"
        )

    @safe_fail
    def command_debug(self, expr, glb, lcl):
        from .debug import SnekTalkDb

        expr = expr.lstrip()
        glb = glb or self.glb
        lcl = lcl or self.lcl
        if expr:
            self.session.queue(
                command="echo", value=f"/debug {expr}", process=False
            )
            result = SnekTalkDb().runeval_step(expr, glb, lcl)
            typ = "statement" if result is None else "expression"
            self.session.queue_result(result, type=typ)

        else:
            exc = self.session.blt.get("$$exc_info", None)
            if exc:
                self.session.queue(
                    command="echo",
                    value=f"Debugging {exc[0].__qualname__}: {exc[1]}",
                    language="text",
                    process=False,
                )
                tb = exc[2]
                SnekTalkDb().interaction(tb.tb_frame, tb)
            else:
                self.session.queue_result(
                    H.div("Last expression was not an exception"),
                    type="exception",
                )

    @safe_fail
    def command_shell(self, expr, glb, lcl):
        expr = expr.lstrip()
        self.session.queue(command="echo", value=f"//{expr}", process=False)
        proc = subprocess.run(expr, shell=True, capture_output=True)
        if proc.stdout:
            self.session.queue_result(
                H.div["snek-shellout"](proc.stdout.decode("utf8")),
                type="expression",
            )
        if proc.stderr:
            self.session.queue_result(
                H.div["snek-shellout"](proc.stderr.decode("utf8")),
                type="exception",
            )

    @safe_fail
    def command_thread(self, expr, glb, lcl):
        assert isinstance(expr, str)
        expr = expr.lstrip()
        self.session.queue(command="echo", value=expr, process=False)

        def run():
            result = self.eval(expr, glb, lcl)
            typ = "statement" if result is None else "expression"
            self.session.blt["_"] = result
            self.session.queue_result(result, type=typ)

        threads.run_in_thread(run, session=current_session())

    def command_quit(self, expr, glb, lcl):
        self.session.queue_result(H.div("/quit"), type="echo")
        self.session.queue_result(
            H.div("Quitting interpreter in ", self.format_modname()),
            type="info",
        )
        # Small delay so that the messages get flushed
        time.sleep(0.01)
        raise StopEvaluator()

    def missing(self, expr, cmd, arg, glb, lcl):
        self.session.queue(command="echo", value=expr, process=False)
        self.session.queue_result(
            H.div(f"Command '{cmd}' does not exist"), type="exception"
        )

    def dispatch(self, expr, glb=None, lcl=None):
        if expr.startswith("//"):
            cmd = "shell"
            arg = expr[2:]
        elif match := cmd_rx.fullmatch(expr):
            cmd, arg = match.groups()
            if arg is None:
                arg = ""
        elif expr.startswith("?"):
            cmd = "dir"
            arg = expr[1:]
        else:
            cmd = "eval"
            arg = expr

        cmd = self.shorthand.get(cmd, cmd)
        method = getattr(self, f"command_{cmd}", None)
        if method is None:
            self.missing(expr, cmd, arg, glb, lcl)
        else:
            method(arg, glb, lcl)

    def loop(self):
        pyv = sys.version_info
        self.session.queue_result(
            H.div(
                "Starting interpreter in ",
                self.format_modname(),
                H.br(),
                f"Snektalk {version} using Python {pyv.major}.{pyv.minor}.{pyv.micro}",
            ),
            type="info",
        )
        try:
            while True:
                try:
                    prompt = H.span["snek-input-mode-python"](self.prompt)
                    with self.session.prompt(prompt) as cmd:
                        if cmd["command"] == "expr":
                            expr = cmd["expr"]
                            if expr.strip():
                                try:
                                    self.dispatch(expr)
                                except SnektalkInterrupt as exc:
                                    print(exc)
                        elif cmd["command"] == "noop":
                            pass
                except SnektalkInterrupt:
                    continue
        except StopEvaluator:
            return

    run = command_eval



import types
from dataclasses import dataclass
from itertools import count, product

from hrepr import H, Interface, standard_terminal
from jurigged import make_recoder
from ovld import OvldMC, exactly, ovld

from ..utils import Interactor, ObjectFields, format_libpath, represents

########
# edit #
########


@ovld.dispatch
def edit(self, obj, **kwargs):
    if hasattr(obj, "__snek_edit__"):
        return obj.__snek_edit__(**kwargs)
    else:
        return self[type(obj)](obj, **kwargs)


@ovld
def edit(
    obj: (type, types.FunctionType, types.CodeType, types.ModuleType), **kwargs
):
    recoder = make_recoder(obj)
    return recoder and SnekRecoder(recoder, obj, **kwargs)


@ovld
def edit(obj: (tuple, int, float, bool, type(None), frozenset), **kwargs):
    raise ValueError(f"Cannot edit data with type {type(obj).__qualname__}")


@ovld
def edit(obj: object, **kwargs):
    return DataEditor(obj, **kwargs)


######################
# Python code editor #
######################


class SnekRecoder(Interactor):

    js_constructor = "LiveEditor"
    js_source = "/scripts/liveedit.js"

    def __init__(
        self,
        recoder,
        fn,
        code_highlight=None,
        max_height=500,
        autofocus=False,
        evaluator=None,
    ):
        self.recoder = recoder
        self.fn = fn
        super().__init__(
            {
                "content": {
                    "live": self.recoder.focus.codestring,
                    "saved": self.recoder.focus.stashed.content,
                },
                "filename": format_libpath(self.recoder.codefile.filename),
                "highlight": code_highlight,
                "max_height": max_height,
                "autofocus": autofocus,
            }
        )
        self.recoder.on_status.register(self.on_status)

    def on_status(self, recoder, status):
        if status == "out-of-sync":
            self.js.setStatus("dirty", "out of sync")

    def py_save(self, new_source):
        self.recoder.patch(new_source)
        return True

    def py_commit(self, new_source):
        if self.py_save(new_source):
            self.recoder.commit()


##########################
# Data structures editor #
##########################


class DataEditor(Interactor):

    js_constructor = "LiveEditor"
    js_source = "/scripts/liveedit.js"

    def __init__(
        self,
        obj,
        *,
        evaluator,
        code_highlight=None,
        max_height=500,
        autofocus=False,
    ):
        self.obj = obj
        self.repr = editable_repr(self.obj, evaluator=evaluator)
        self.repr = self.repr.replace(" \n", "\n")
        self.model = evaluator.eval(self.repr)
        self.evaluator = evaluator
        super().__init__(
            {
                "content": {"live": self.repr, "saved": self.repr,},
                "filename": "<data>",
                "highlight": code_highlight,
                "max_height": max_height,
                "autofocus": autofocus,
            }
        )

    def py_save(self, new_source):
        new_model = self.evaluator.eval(new_source)
        _merge(self.obj, self.model, new_model)

    def py_commit(self, new_source):
        raise Exception("This data does not correspond to a file")


class EditableRepr(metaclass=OvldMC):
    def __init__(self, max_depth, evaluator):
        self.max_depth = max_depth
        self.evaluator = evaluator
        self.seen = set()

    def ref(self, obj, wrap=H.atom):
        paths = []

        if type(obj) in (str, type(None), bool, int, float):
            return wrap(repr(obj))

        if (mod := getattr(obj, "__module__", None)) and (
            name := getattr(obj, "__qualname__", None)
        ):
            paths.append(name)
            paths.append(f"{mod}.{name}")
            paths.append(f"sktk.imp.{mod}.{name}")

        for path in paths:
            try:
                result = self.evaluator.eval(path)
                if result is obj:
                    return wrap(path)
            except Exception as exc:
                pass
        else:
            return wrap(self.evaluator.session.getvar(obj))

    def run(self, obj, *, depth):
        oid = id(obj)
        if oid in self.seen:
            return self.ref(obj)
        result = self._run(obj, depth=depth)
        self.seen.add(oid)
        return result

    @ovld
    def _run(
        self,
        s: (
            exactly(str),
            exactly(type(None)),
            exactly(bool),
            exactly(int),
            exactly(float),
        ),
        *,
        depth,
    ):
        return H.atom(repr(s))

    @ovld
    def _run(self, s: (types.FunctionType, types.ModuleType, type), *, depth):
        return self.ref(s)

    @ovld
    def _run(self, d: exactly(dict), *, depth):
        if depth == self.max_depth:
            return self.ref(d)

        items = list(d.items())
        return H.bracketed(
            [
                H.pair(
                    self.run(k, depth=depth + 1),
                    self.run(v, depth=depth + 1),
                    delimiter=": ",
                )
                for k, v in items
            ],
            start="{",
            end="}",
            delimiter=", ",
        )

    @ovld
    def _run(self, xs: exactly(list), *, depth):
        if depth == self.max_depth:
            return self.ref(xs)

        return H.bracketed(
            [self.run(x, depth=depth + 1) for x in xs],
            start="[",
            end="]",
            delimiter=", ",
        )

    @ovld
    def _run(self, xs: exactly(tuple), *, depth):
        if depth == self.max_depth:
            return self.ref(xs)

        elems = [self.run(x, depth=depth + 1) for x in xs]
        if len(xs) == 1:
            elems.append(H.atom(""))
        return H.bracketed(elems, start="(", end=")", delimiter=", ")

    @ovld
    def _run(self, xs: exactly(set), *, depth):
        if depth == self.max_depth:
            return self.ref(xs)

        if not xs:
            return H.atom("set()")
        else:
            return H.bracketed(
                [self.run(x, depth=depth + 1) for x in xs],
                start="{",
                end="}",
                delimiter=",",
            )

    @ovld
    def _run(self, obj: object, *, depth):
        d = getattr(obj, "__dict__", None)
        if depth == self.max_depth or not isinstance(d, dict):
            return self.ref(obj)

        call = f"sktk.mod({self.ref(type(obj), wrap=str)})"
        return H.bracketed(
            [
                H.pair(k, self.run(v, depth=depth + 1), delimiter="=")
                for k, v in d.items()
            ],
            start=f"{call}(",
            end=")",
            delimiter=", ",
        )

        return self.ref(obj)

    def __call__(self, obj):
        return self.run(obj, depth=0)


def editable_repr(obj, max_depth=2, evaluator=None):
    er = EditableRepr(max_depth=max_depth, evaluator=evaluator)
    spec = standard_terminal(er(obj))
    return spec.to_string(max_col=80)


@ovld.dispatch(initial_state=lambda: {"seen": set()})
def _merge(self, obj, m1, m2):
    if m1 == m2:
        return obj
    elif type(m1) is not type(m2):
        if isinstance(m2, ObjectFields):
            raise ValueError("Cannot merge ObjectFields object")
        return m2
    elif id(obj) in self.seen:
        return obj
    else:
        self.seen.add(id(obj))
        return self.call(obj, m1, m2)


@ovld
def _merge(self, obj: dict, m1: dict, m2):
    keys = {*m1.keys(), *m2.keys()}
    for k in keys:
        if k not in m2:
            del obj[k]
        elif k not in m1:
            obj[k] = m2[k]
        else:
            obj[k] = self(obj[k], m1[k], m2[k])
    return obj


@ovld
def _merge(self, obj: list, m1: list, m2):
    ops = _sequence_operations(obj, m1, m2)
    obj[:] = []
    for op, *args in ops:
        if op == "merge":
            value = self(*args)
        else:
            (value,) = args
        obj.append(value)
    return obj


@ovld
def _merge(
    self,
    obj: (
        exactly(str),
        exactly(type(None)),
        exactly(bool),
        exactly(int),
        exactly(float),
    ),
    m1,
    m2,
):
    return m2


@ovld
def _merge(self, obj: object, m1: ObjectFields, m2):
    if m1.cls is not m2.cls:
        raise TypeError("Cannot update object to a different type")
    self(obj.__dict__, m1.fields, m2.fields)
    return obj


def _sequence_operations(obj, m1, m2):
    used1 = set()
    fits = [
        (_correspondence(x, y), -abs(i - j), -i, (i, j))
        for (i, x), (j, y) in product(enumerate(m1), enumerate(m2))
    ]
    fits.sort(reverse=True)
    operations = [("set", y) for y in m2]
    for _, _, _, (i, j) in fits:
        if j not in operations and i not in used1:
            operations[j] = ("merge", obj[i], m1[i], m2[j])
            used1.add(i)
    return operations


def _correspondence(x, y):
    if x is y:
        return 3
    elif x == y:
        return 2
    elif type(x) is type(y):
        return 1
    else:
        return 0



import bdb
import os
import re

from hrepr import H
from jurigged import registry

from .evaluator import Evaluator
from .session import current_session
from .utils import ReadOnly, format_libpath

_here = os.path.dirname(__file__)


class SnekTalkDb(bdb.Bdb):
    # Note: Many methods have been lifted from pdb.py in the standard lib
    # and slightly adapted

    def reset(self):
        super().reset()
        self.session = current_session()
        self.prompt = H.span["snek-input-mode-debug"]("debug>")
        self.nav = ReadOnly(
            {"filename": "", "content": "", "firstlineno": 1, "highlight": 1}
        )
        self.session.set_nav(self.nav)
        self.evaluator = Evaluator(None, {}, {}, self.session)
        self.last_method = (lambda *_: None), None

    def user_line(self, frame):
        if frame.f_code.co_filename.startswith(_here):
            # Avoid running the debugger in snektalk code
            self.set_continue()
        elif getattr(self, "step_now", False):
            self.step_now = False
            self.set_step()
        else:
            self.set_frame(frame)

    def show_frame(self, frame):
        try:
            cf, defn = registry.find(frame.f_code)
        except Exception as exc:
            cf, defn = None, None
        if defn is None:
            self.session.add_nav_action(
                "update",
                lambda: self.nav.js.update(
                    f"'Could not find source code for {frame.f_code.co_name}'"
                ),
            )
        else:
            self.session.add_nav_action(
                "update",
                lambda: self.nav.js.update(
                    defn.codestring,
                    format_libpath(defn.filename),
                    defn.stashed.lineno,
                    frame.f_lineno,
                ),
            )

    def get_frame(self):
        return self.stack[self.current][0]

    def set_frame(self, frame, tb=None):
        self.stack, self.current = self.get_stack(frame, tb)
        self.proceed = False
        while not self.proceed:
            self.show_frame(self.get_frame())
            with self.session.prompt(prompt=self.prompt, nav=self.nav) as cmd:
                if cmd["command"] == "expr":
                    expr = cmd["expr"]
                    self.do(expr)
                elif cmd["command"] == "noop":
                    pass

    def error(self, message):
        print(H.i(message, style="color:red"))

    def do(self, code):
        frame = self.get_frame()
        gs = frame.f_globals
        ls = frame.f_locals
        cmd = None
        if not code.strip():
            method, arg = self.last_method
        else:
            m = re.match(r"^[?!]+", code)
            if m:
                cmd = code[: m.end()]
                arg = code[m.end() :]
            elif " " in code:
                cmd, arg = code.split(" ", 1)
            else:
                cmd = code
                arg = ""
            method = self.__commands__.get(cmd, None)
        if method and cmd != "!":
            self.last_method = method, arg
        if method:
            return method(self, arg, gs, ls)
        else:
            return self.command_print(code, gs, ls)

    def command_print(self, code, gs, ls):
        """<b>p(rint) expression</b>

        Print the value of the expression. Synonyms: !, pp
        """
        self.evaluator.run(code, gs, ls)

    def command_step(self, arg, gs, ls):
        """<b>s(tep)</b>

        Execute the current line, stop at the first possible occasion
        (either in a function that is called or in the current function).
        """
        self.set_step()
        self.proceed = True

    def command_next(self, arg, gs, ls):
        """<b>n(ext)</b>

        Continue execution until the next line in the current function
        is reached or it returns.
        """
        self.set_next(self.get_frame())
        self.proceed = True

    def command_continue(self, arg, gs, ls):
        """<b>c(ont(inue))</b>

        Continue execution, only stop when a breakpoint is encountered.
        """
        self.set_continue()
        self.proceed = True

    def command_return(self, arg, gs, ls):
        """<b>r(eturn)</b>

        Continue execution until the current function returns.
        """
        self.set_return()
        self.proceed = True

    def command_up(self, arg, gs, ls):
        """<b>u(p) [count]</b>

        Move the current frame count (default one) levels up in the
        stack trace (to an older frame).
        """
        if self.current == 0:
            self.error("Oldest frame")
            return
        try:
            count = int(arg or 1)
        except ValueError:
            self.error("Invalid frame count (%s)" % arg)
            return
        if count < 0:
            self.current = 0
        else:
            self.current = max(0, self.current - count)

    def command_down(self, arg, gs, ls):
        """<b>d(own) [count]</b>

        Move the current frame count (default one) levels down in the
        stack trace (to a newer frame).
        """
        if self.current + 1 == len(self.stack):
            self.error("Newest frame")
            return
        try:
            count = int(arg or 1)
        except ValueError:
            self.error("Invalid frame count (%s)" % arg)
            return
        if count < 0:
            self.current = len(self.stack) - 1
        else:
            self.current = min(len(self.stack) - 1, self.current + count)

    def command_help(self, arg, gs, ls):
        """<b>h(elp)</b>

        Without argument, print the list of available commands.
        With a command name as argument, print help about that command.
        """
        if arg:
            html = H.div(style="border:1px solid black; padding: 5px")
            self.session.queue(
                command="result",
                type="print",
                value=html(H.raw(self.__commands__[arg].__doc__)),
            )
        else:
            commands = [x.__doc__ for x in set(self.__commands__.values())]
            commands.sort()
            html = H.div(style="border:1px solid black; padding: 5px")
            for cmd in commands:
                html = html(H.div(H.raw(cmd)))
            self.session.queue(
                command="result", type="print", value=html,
            )

    def command_quit(self, arg, gs, ls):
        """<b>q(uit)</b>

        Quit from the debugger. The program being executed is aborted.
        """
        self._user_requested_quit = True
        self.set_quit()
        self.proceed = True

    def runeval_step(self, expr, glb, lcl):
        self.step_now = True
        return self.runeval(expr, glb, lcl)

    def interaction(self, frame, tb):
        self.reset()
        self.set_frame(frame, tb)

    __commands__ = {
        "step": command_step,
        "s": command_step,
        "next": command_next,
        "n": command_next,
        "continue": command_continue,
        "c": command_continue,
        "return": command_return,
        "r": command_return,
        "up": command_up,
        "u": command_up,
        "down": command_down,
        "d": command_down,
        "print": command_print,
        "p": command_print,
        "pp": command_print,
        "!": command_print,
        "help": command_help,
        "h": command_help,
        "?": command_help,
        "quit": command_quit,
        "q": command_quit,
    }



import weakref
from collections import deque
from itertools import count

_c = count()


class UNAVAILABLE:
    def __init__(self):
        pass

    def __hrepr__(self, H, hrepr):
        return H.span("<UNAVAILABLE>", style="color:#bbb")


UNAVAILABLE = UNAVAILABLE()


class CallbackRegistry:
    def __init__(self, keep=10000):
        self.keep = keep
        self.id = 0
        self.weak_map = {}
        self.strong_map = {}
        self.strong_ids = deque()

    def cleanup(self, id):
        del self.weak_map[id]
        del self.callbacks[id]

    def register(self, method):
        self.id += 1
        currid = self.id
        try:
            self.weak_map[currid] = weakref.WeakMethod(method)
        except TypeError:
            self.strong_map[currid] = method
            self.strong_ids.append(currid)
            if len(self.strong_ids) > self.keep >= 0:
                rm = self.strong_ids.popleft()
                del self.strong_map[rm]
        return currid

    def resolve(self, id):
        try:
            m = self.weak_map[id]()
        except KeyError:
            m = self.strong_map.get(id, None)

        if m is None:
            raise KeyError(id)

        return m


callback_registry = CallbackRegistry()



import atexit
import errno
import ipaddress
import json
import os
import random
import socket
import subprocess
import time
import urllib
import webbrowser


def create_socket(socket_path):
    def rmsock():
        os.remove(socket_path)
        os.remove(f"{socket_path}.host")

    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    s.bind(socket_path)
    usr = os.environ["USER"]
    host = socket.gethostname()
    with open(f"{socket_path}.host", "w") as hostfile:
        hostfile.write(f"{usr}@{host}")
    atexit.register(rmsock)
    return s


def create_inet(randomize=False):
    # Generate a random loopback address (127.x.x.x)
    if randomize:
        addr = ipaddress.IPv4Address("127.0.0.1") + random.randrange(
            2 ** 24 - 2
        )
        addr = str(addr)
    else:
        addr = "localhost"
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.bind((addr, 0))
    except OSError as exc:
        if exc.errno == errno.EADDRNOTAVAIL:
            # The full 127.x.x.x range may not be available on this system
            sock.bind(("localhost", 0))
    return sock


def connect_to_existing(addr, port, sock):
    assert sock is not None
    if ":" in addr:
        addr, jump_port = addr.split(":")
    else:
        jump_port = None

    print("Collecting information about remote socket")
    cmd = [
        "ssh",
        addr,
        *(["-p", jump_port] if jump_port else []),
        f"pwd; echo $USER; hostname; cat {sock}.host",
    ]
    output = subprocess.check_output(cmd)
    home, hostuser, hostname, sockhost = output.decode("utf8").split("\n")[:4]
    host = f"{hostuser}@{hostname}"
    if sock.startswith("/"):
        location = sock
    else:
        location = f"{home}/{sock}"

    print(f"Socket host: {sockhost}")
    print(f"Socket path: {location}")

    if port is None:
        sock = create_inet()
        localhost, port = sock.getsockname()
        sock.close()
    else:
        localhost = "localhost"

    print(f"Forwarding to local port {port}")

    if host == sockhost:
        cmd = [
            "ssh",
            "-nNCL",
            f"{localhost}:{port}:{location}",
            addr,
            "-p",
            jump_port,
        ]

    else:
        cmd = [
            "ssh",
            "-nNCL",
            f"{localhost}:{port}:{location}",
            "-J",
            f"{addr}:{jump_port}" if jump_port else addr,
            sockhost,
        ]

    print("Run:", " ".join(cmd))
    proc = subprocess.Popen(cmd)
    url = f"http://{localhost}:{port}/"
    print(f"Connecting to {url}")

    wait_time = 0.05
    for i in range(5):
        time.sleep(wait_time)
        wait_time *= 2
        try:
            with urllib.request.urlopen(f"{url}status") as req:
                status = json.loads(req.read().decode("utf8"))
                if status["status"] == "OK":
                    break
        except urllib.error.URLError:
            continue

    webbrowser.open(url)
    proc.wait()



import os
from itertools import count
from types import FunctionType, MethodType

from hrepr import H, hjson, hrepr

from .registry import callback_registry
from .session import current_session

_count = count()


##########################
# Special JSON converter #
##########################


@hjson.dump.variant
def _sktk_hjson(self, fn: (MethodType, FunctionType)):
    method_id = callback_registry.register(fn)
    return f"$$SKTK({method_id})"


def sktk_hjson(obj):
    return str(_sktk_hjson(obj))


#############
# Utilities #
#############


def join(elems, sep):
    rval = [elems[0]]
    for elem in elems[1:]:
        rval.append(sep)
        rval.append(elem)
    return rval


_path_replacements = {
    "PWD": "",
    "CONDA_PREFIX": "$CONDA_PREFIX/",
    "VIRTUAL_ENV": "$VIRTUAL_ENV/",
    "HOME": "~/",
}


def format_libpath(path):
    for var, pfx in sorted(_path_replacements.items(), key=lambda kv: kv[1]):
        if (val := os.environ.get(var, None)) is not None:
            if not val.endswith("/"):
                val += "/"
            if path.startswith(val):
                return os.path.join(pfx, path[len(val) :])
    else:
        return path


class Importer:
    def __getattr__(self, module_name):
        return __import__(module_name)


class ObjectFields:
    def __init__(self, cls, **fields):
        self.cls = cls
        self.fields = fields

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and self.cls == other.cls
            and self.fields == other.fields
        )


def mod(cls):
    def f(**fields):
        return ObjectFields(cls, **fields)

    return f


###########################
# Click/shift-click logic #
###########################


def pastevar(obj):
    sess = current_session()
    sess.queue(command="pastevar", value=sess.getvar(obj))


def pastecode(code):
    sess = current_session()
    sess.queue(command="pastevar", value=code)


def _default_click(obj, evt):
    if evt.get("shiftKey", False):
        sess = current_session()
        sess.queue(command="result", value=hrepr(obj), type="print")
    else:
        pastevar(obj)


def _safe_set(elem, **props):
    if elem.is_virtual():
        return H.div(elem, **props)
    else:
        return elem(**props)


def represents(obj, elem, pinnable=False):
    if obj is None:
        return elem
    elif elem.get_attribute("objid", None) is not None:
        return _safe_set(elem, pinnable=pinnable)
    else:
        method_id = callback_registry.register(MethodType(_default_click, obj))
        return _safe_set(elem, objid=method_id, pinnable=pinnable)


##############
# Interactor #
##############


class BaseJSCaller:
    def __init__(self, interactor, jsid):
        self._interactor = interactor
        self._jsid = jsid
        self._session = current_session()

    def _getcode(self, method_name, args):
        if not self._interactor:
            raise Exception("The JavaScript interface is not active.")
        argtext = ",".join(map(sktk_hjson, args))
        return f"""
        require(
            ['{self._jsid}'],
            wobj => {{
                let obj = wobj.deref();
                if (obj !== null) {{
                    obj.{method_name}({argtext});
                }}
            }}
        );
        """


class AJSCaller(BaseJSCaller):
    # TODO
    def __getattr__(self, method_name):
        async def call(*args):
            code = self._getcode(method_name, args)
            prom = asyncio.Promise()
            self._session.queue(command="eval", value=code, promise=prom)
            return await prom

        return call


class JSCaller(BaseJSCaller):
    def __init__(self, interactor, jsid, return_hrepr):
        super().__init__(interactor, jsid)
        self._return_hrepr = return_hrepr

    def __getattr__(self, method_name):
        def call(*args):
            code = self._getcode(method_name, args)
            if self._return_hrepr:
                return H.javascript(code)
            else:
                self._session.queue(command="eval", value=code)

        return call


class Interactor:
    js_requires = {}
    js_code = None

    @classmethod
    def show(cls, *args, nav=False, **kwargs):
        instance = cls(*args, **kwargs)
        html = hrepr(instance)
        if nav:
            current_session().queue(command="set_nav", value=html)
        else:
            print(html)
        return instance

    def __init__(self, parameters):
        self.jsid = f"interactor{next(_count)}"
        self.js = JSCaller(self, self.jsid, return_hrepr=False)
        self.hjs = JSCaller(self, self.jsid, return_hrepr=True)
        self.ajs = AJSCaller(self, self.jsid)
        self.parameters = parameters
        methods = {}
        for method_name in dir(self):
            if method_name.startswith("py_"):
                methods[method_name[3:]] = getattr(self, method_name)
        self.parameters["py"] = methods
        self.active = False

    def __bool__(self):
        return self.active

    @classmethod
    def __hrepr_resources__(cls, H):
        js_requires = {
            name: {"url": x, "varname": name} if isinstance(x, str) else x
            for name, x in cls.js_requires.items()
        }

        reqs = [
            H.javascript(export=name, src=data["url"])
            for name, data in js_requires.items()
        ]
        if cls.js_code:
            main = H.javascript(
                cls.js_code,
                require={
                    name: data["varname"] for name, data in js_requires.items()
                },
            )
        else:
            main = H.javascript(src=cls.js_source)
        return [*reqs, main(export=cls.js_constructor)]

    def __hrepr__(self, H, hrepr):
        params = sktk_hjson(self.parameters)
        self.active = True
        tmpid = f"$$tmp{next(_count)}"
        return H.div(
            H.div(id=tmpid),
            H.script(
                f"""
                (() => {{
                    let elem = document.getElementById('{tmpid}');
                    let existing = document.getElementById('{self.jsid}');
                    if (existing && existing.handler) {{
                        // Move the existing div
                        elem.parentElement.replaceChild(
                            existing, elem
                        );
                    }}
                    else {{
                        elem.id = '{self.jsid}';
                        requirejs.undef('{self.jsid}');
                        define(
                            '{self.jsid}',
                            ['{self.js_constructor}'],
                            ctor => {{
                                let obj = new ctor(elem, {params});
                                elem.handler = obj;
                                return new WeakRef(obj);
                            }}
                        );
                        require(['{self.jsid}'], _ => null);
                    }}
                }})();
                """
            ),
        )


####################
# Misc interactors #
####################


class ReadOnly(Interactor):
    js_constructor = "ReadOnlyEditor"
    js_source = "/scripts/readonly.js"



import sys
from types import SimpleNamespace

from coleo import Option, default, run_cli
from jurigged import registry
from jurigged.live import find_runner
from jurigged.utils import glob_filter

from .evaluator import Evaluator, threads
from .network import connect_to_existing
from .server import serve


def main():
    mod, run, sess, thread = run_cli(cli)

    try:
        if run is not None:
            if thread:
                threads.run_in_thread(run, session=sess)
            else:
                run()
    except Exception as exc:
        if sess is not None:
            sess.blt["$$exc_info"] = sys.exc_info()
            sess.queue_result(exc, type="exception")
        else:
            raise
    finally:
        Evaluator(mod, vars(mod) if mod else {}, None, sess).loop()


def cli():
    # Module or module:function to run
    # [options: -m]
    module: Option = default(None)

    # Path to the script to run
    # [positional: ?]
    script: Option = default(None)

    # Script arguments
    # [remainder]
    argv: Option

    # Don't watch changes on the filesystem
    # [options: --no-watch]
    no_watch: Option & bool = default(False)

    # Server port
    # [alias: -p]
    port: Option & int = default(None)

    # Path to socket
    # [alias: -S]
    socket: Option & str = default(None)

    # Hostname to connect to an existing instance
    # [alias: -c]
    connect: Option & str = default(False)

    # Run the program in a thread
    # [alias: -t]
    thread: Option & bool = default(False)

    # Show the version
    version: Option & bool = default(False)

    if version:
        from .version import version

        print(version)
        return

    if connect:
        connect_to_existing(connect, port, socket)
        return

    pattern = glob_filter(".")
    if no_watch:
        watch_args = None
    else:
        watch_args = {
            "registry": registry,
            "pattern": pattern,
        }

    server_args = {
        "port": port,
        "sock": socket,
    }

    sess = serve(
        watch_args=watch_args,
        template={"title": module or script or "snektalk"},
        **server_args,
    )

    opts = SimpleNamespace(module=module, script=script, rest=argv,)

    mod, run = find_runner(opts, pattern)
    return mod, run, sess, thread



import json
import os


def get_config_path(name=None):
    path = os.path.expanduser("~/.config/snektalk")
    if name is not None:
        path = os.path.join(path, name)
    return path


def mayread(path, default=None):
    if not os.path.exists(path):
        return default
    with open(path) as f:
        try:
            return json.load(f)
        except json.JSONDecodeError:
            return default


def _dump(contents, f):
    json.dump(contents, f, indent=4)


def maywrite(path, contents, dump=_dump):
    cdir = os.path.dirname(path)
    os.makedirs(cdir, exist_ok=True)
    with open(path, "w") as f:
        _dump(contents, f)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['snektalk', 'snektalk.feat']

package_data = \
{'': ['*'], 'snektalk': ['assets/*', 'assets/scripts/*', 'assets/style/*']}

install_requires = \
['coleo>=0.2.2,<0.3.0',
 'hrepr>=0.3.11,<0.4.0',
 'jurigged>=0.3.1,<0.4.0',
 'ovld>=0.2.10,<0.3.0',
 'ptera>=0.3.3,<0.4.0',
 'sanic>=20.9.1,<21.0.0']

entry_points = \
{'console_scripts': ['sktk = snektalk.cli:main',
                     'snektalk = snektalk.cli:main']}

setup_kwargs = {
    'name': 'snektalk',
    'version': '0.0.12',
    'description': 'Advanced Python REPL',
    'long_description': '\n**Note:** This project is at a very early stage and no aspect of it should be considered stable before v0.1.0 is released.\n\n# S N E K T A L K\n\nSnektalk is a groundbreaking new kind of REPL.\n\n* Live code editing\n* Rich and interactive object representations\n* Built-in debugger\n* Connect to remote processes\n* Not another Jupyter clone\n\n# Install\n\n```bash\npip install snektalk\n```\n\n# Usage\n\nYou can simply use `snektalk` instead of `python` to run a script.\n\n```\nusage: snektalk [-h] [--connect VALUE] [-m VALUE] [--no-watch] [--port NUM]\n                [--socket VALUE] [--thread] [--version]\n                [SCRIPT] ...\n\npositional arguments:\n  SCRIPT                Path to the script to run\n  ARGV                  Script arguments\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --connect VALUE, -c VALUE\n                        Hostname to connect to an existing instance\n  -m VALUE              Module or module:function to run\n  --no-watch            Don\'t watch changes on the filesystem\n  --port NUM, -p NUM    Server port\n  --socket VALUE, -S VALUE\n                        Path to socket\n  --thread, -t          Run the program in a thread\n  --version             Show the version\n```\n\n# Features\n\nAt a glance Snektalk might appear similar to Jupyter notebooks, but it follows different paradigms. It has no "cells" and is meant to be used like a straightforward REPL or command line. At the same time, it has many features neither standard REPLs nor Jupyter tend to have.\n\n## Edit functions and data\n\nSimply type `/edit func` and you will be greeted with a small inline editor for the source code of `func`. You may change it and hit `Ctrl+Enter` to change it in the current process, or `Ctrl+Shift+Enter` to save it back into the original file it came from. You can come back to it at any time, of course.\n\n![edit](https://user-images.githubusercontent.com/599820/116953136-7c74de00-ac5a-11eb-9868-a53da72a1f5d.gif)\n\nVirtually *any* function can be edited, whether it is yours or comes from a third party library or even the standard library.\n\n`/edit` also works on data structures. You will be given an editable sandbox where you can change dictionaries, reorder lists, change the values of the fields of an object, and so on. Objects can even define a custom `__snek_edit__` method to control how they are edited.\n\n![edit-data](https://user-images.githubusercontent.com/599820/116953144-826abf00-ac5a-11eb-93b7-cfdcc46ac166.gif)\n\n## Rich and interactive representations\n\nSnektalk does not print lists, dictionaries or objects as mere text, but as rich HTML objects using [hrepr](https://github.com/breuleux/hrepr).\n\n![repr](https://user-images.githubusercontent.com/599820/116953095-5ea77900-ac5a-11eb-8091-5b27a3a795dc.gif)\n\n`Ctrl+Click` (or `Cmd+Click` on Mac) the representation of an object to put it in a temporary variable. This makes it very easy to test or play with objects that are deeply nested in another.\n\n![click](https://user-images.githubusercontent.com/599820/116953201-ac23e600-ac5a-11eb-9464-3aeccc9632d0.gif)\n\nRepresentations are highly customizable and recursive representations can be defined and configured in a snap. See [here](https://github.com/breuleux/hrepr#custom-representations) for how to define custom representations.\n\nThe representation of exceptions is particularly interesting because each frame is associated to a live editor, so you can simply fix the error right there as you see it.\n\n![exc](https://user-images.githubusercontent.com/599820/116953211-b34af400-ac5a-11eb-9d7e-37b51a7e955f.gif)\n\n## Visualization\n\nSnektalk supports elaborate visualizations: plots, graphs, and so on. Integrating a new or existing JavaScript library is mostly a matter of linking it from a CDN and writing a small wrapper.\n\n![plot](https://user-images.githubusercontent.com/599820/116953224-ba720200-ac5a-11eb-8a4e-aba17fb214bc.gif)\n\nIt is also easy to configure visualizations so that various interactions call Python callbacks. One great use of this feature is the ability to click on nodes or points in a graph to put the underlying data in a variable and paste it into the REPL\'s input box so that you can analyze it further.\n\n![graph](https://user-images.githubusercontent.com/599820/116955206-e2b02f80-ac5f-11eb-8474-dbdcd59cdd3a.gif)\n\n## Debugging\n\n`/debug f(x, y)` will enter a function call in debugger mode. Snektalk\'s debugger is quite similar to `pdb` and the usual `pdb` commands (`step`, `next`, `continue`, etc.) should work just the same.\n\n![debug](https://user-images.githubusercontent.com/599820/116955224-eba10100-ac5f-11eb-81a8-4b042718611b.gif)\n\n## Threads\n\n`/thread f(x, y)` will run `f(x, y)` in a separate thread, which lets you keep working while it\'s running. Each thread is given a mnemonic name so that you can easily `/kill` them.\n\n![thread](https://user-images.githubusercontent.com/599820/116955232-f0fe4b80-ac5f-11eb-8578-079f5e753052.gif)\n\nYou may use `snektalk -t` to start the main script in a thread, giving you immediate access to the REPL. This will allow you to inspect or fiddle with the global state while the script is running, among other things.\n\n# Using on a remote\n\nYou can run `snektalk` on a remote computer and connect to it over SSH. To do so, you will need to run two commands:\n\n**On remote side:**\n\nRun `snektalk` on the remote side with the `-S` option to connect to a UNIX socket on the filesystem. The interface will be served through that socket instead of a port.\n\n```bash\nuser@remote$ snektalk -S ~/sock/script.sock script.py\n```\n\n**On local computer:**\n\nOnce the remote process is running, run `snektalk` with the `-c` option to specify which host to connect to, and `-S` pointing to the socket file.\n\n```bash\nme@local$ snektalk -c user@remote -S sock/script.sock\n```\n\nThis will work for the whole duration of the remote process (note that the local Snektalk invocation doesn\'t do much more than invoke the right SSH command to forward the remote socket to a local port).\n\n**Note about compute nodes:** In order to facilitate use on clusters where the compute nodes may only be available through a connection to the login node, Snektalk will store the hostname in a separate file (`sock/script.host` in the above example) and will attempt to connect to it automatically. Therefore, you should be able to run `snektalk` on a compute node, then give the login node as the argument to `-c`, and Snektalk will use the login node as a jump host to connect to the right compute node.\n\n# Commands\n\n* `/debug expr` -- Debug an expression\n* `/dir expr` -- List all members of the object returned by the expression\n  * `?expr` -- Same as `/dir expr`\n* `/edit expr` -- Open an editor for the object returned by the expression\n* `/quit` -- Quit Snektalk\n* `/shell command` -- Run shell command\n  * `//command` -- Same as `/shell command`\n* `/status` -- List all the status messages received so far\n\n## Thread-related commands\n\n* `/attach thread` -- Switch to the REPL of the named thread (you can also click on the prompt to list the threads, then click on the one you want to attach)\n* `/detach` -- Undo last /attach\n* `/kill thread` -- Stop a named thread\n* `/thread expr` -- Run expression in new thread\n\n# Keyboard and mouse bindings\n\nNote: in what follows, `Cmd+X` means `Cmd+X` on MacOS and `Ctrl+X` on other platforms, unless otherwise specified.\n\n## Global bindings\n\n* `Cmd+P` -- Focus the REPL\n* `Cmd+Click` -- Put object in a variable\n* `Cmd+Alt+Click` -- Pin to the side\n\n## Repl\n\n* `Shift+Enter` -- Add new line in REPL without submitting\n* `Ctrl+Enter` -- Submit\n* `Up/Down` -- Go up/down in history (filtered by contents of REPL)\n* `Ctrl+R` -- Open history popup (fuzzy search)\n  * Note: `Shift-Up/Down` when history popup is active will select multiple entries\n* `Ctrl+L` -- Clear all scrollback\n* `Ctrl+C` (MacOS) -- Interrupt current command\n* `Cmd+K Cmd+K` -- Interrupt current command (that\'s `Cmd+K`, twice)\n* `Cmd+B`, `Cmd+Shift+B` -- Cycle through visible editors and REPL\n\n## Function editor\n\n* `Ctrl+Enter` -- Make live and focus REPL\n* `Ctrl+Shift+Enter` -- Make live, commit to file, and focus REPL\n* `Cmd+S` -- Make live\n* `Cmd+Shift+S` -- Make live and commit to file\n* `Cmd+B`, `Cmd+Shift+B` -- Cycle through visible editors and REPL\n\n# Misc\n\n* The status bar at the bottom can be clicked to view a list of all events\n* Click the prompt to reveal a thread selector; if you put a breakpoint in a function called by a different thread, you need to switch to that thread to interact with the breakpoint.\n',
    'author': 'Olivier Breuleux',
    'author_email': 'breuleux@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/breuleux/snektalk',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import builtins
import os
import types
from types import FunctionType, MethodType, ModuleType, SimpleNamespace

from hrepr import H, Hrepr, Tag, hrepr, standard_html
from hrepr.std import _extract_as

from .debug import SnekTalkDb
from .feat.edit import edit
from .session import current_print_session, current_session
from .utils import (
    Importer,
    format_libpath,
    join,
    mod,
    pastevar,
    represents,
    sktk_hjson,
)

##################
# SnekTalk print #
##################


here = os.path.dirname(__file__)
orig_print = print


class PrintSequence(tuple):
    def __hrepr__(self, H, hrepr):
        return H.div["snek-print-sequence"](
            *[H.div(hrepr(x)) for x in self], onclick=False
        )


def snekprint(*args, toplevel=False, to=None, **kwargs):
    sess = current_print_session()
    if sess is None:
        orig_print(*args)
    elif to is not None:
        if toplevel:
            html = hrepr(*args, **kwargs)
        else:
            html = hrepr(PrintSequence(args), **kwargs)
        sess.queue(
            command="insert", value=html, target=to,
        )
    else:
        if all(isinstance(arg, str) for arg in args):
            html = H.div["snek-print-str"](" ".join(args))
        elif toplevel:
            html = hrepr(*args, **kwargs)
        else:
            html = hrepr(PrintSequence(args), **kwargs)
        sess.queue(
            command="result", value=html, type="print",
        )


def snekprint_override(*args, toplevel=False, file=None, std=False, **kwargs):
    builtins.print = orig_print
    sess = current_print_session()
    if sess is None or std or file is not None:
        orig_print(*args, file=file, **kwargs)
    else:
        snekprint(*args, toplevel=toplevel, **kwargs)
    builtins.print = snekprint_override


def insert_at(target, value, index=None, **kwargs):
    html = hrepr(value, **kwargs)
    sess = current_session()
    sess.queue(command="insert", value=html, target=target, index=index)


def fill_at(target, value, **kwargs):
    html = hrepr(value, **kwargs)
    sess = current_session()
    sess.queue(command="fill", value=html, target=target)


def clear_at(target):
    sess = current_session()
    sess.queue(command="clear", target=target)


def help_placeholder(*args, **kwargs):
    print(
        H.span("The help command is not available in snektalk for the moment.")
    )


class HDir:
    exclusions = {"__doc__", "__dict__"}

    def __init__(self, obj):
        self.obj = obj

    def items(self):
        for k in dir(self.obj):
            if k in self.exclusions:
                continue
            try:
                v = getattr(self.obj, k)
            except Exception as e:
                v = e
            yield (k, v)

    def __hrepr__(self, H, hrepr):
        return H.table["hrepr-body"](
            *[H.tr(H.td(k), H.td(" = ", hrepr(v))) for k, v in self.items()]
        )


def hdir(obj):
    return hrepr(HDir(obj), max_depth=2)


###############################
# Add default onclick handler #
###############################


def wrap_onclick(elem, obj, hrepr):
    if not isinstance(obj, Tag) and hrepr.config.interactive is not False:
        return represents(obj, elem, pinnable=True)
    else:
        return elem


def shortname(obj):
    return getattr(obj, "__name__", f"<{type(obj).__name__}>")


class SnekTalkHrepr(Hrepr):
    def collapsible(self, title, body, start_visible=False):
        body = self.H.div(self(body))
        if not start_visible:
            body = body["snek-hidden"]
        return self.H.div["snek-collapsible"](
            self.H.div["snek-collapsible-title"](
                title,
                onclick="this.nextSibling.classList.toggle('snek-hidden')",
            ),
            body,
        )

    def hrepr(self, exc: Exception):
        exc_proper = self.H.div["hrepr-error", "hrepr-instance", "hreprl-h"](
            self.H.div["hrepr-title"](type(exc).__name__),
            self.H.div["hrepr-error-message"](self.H.code(str(exc))),
        )

        if self.state.depth > 0:
            return exc_proper

        tb = exc.__traceback__
        skipping = not self.config.include_snektalk_frames
        parts = []
        curr = tb
        while curr:
            fr = curr.tb_frame
            code = fr.f_code
            filename = code.co_filename
            if filename.startswith(here) and skipping:
                curr = curr.tb_next
                continue
            skipping = False
            if filename.startswith(os.getcwd()) or filename.startswith("<"):
                importance_class = "snek-exception-local"
            else:
                importance_class = "snek-exception-lib"

            hl = curr.tb_lineno - code.co_firstlineno
            try:
                ed = edit(code, code_highlight=hl, max_height=19 * 7)
            except Exception as exc:
                ed = None
            parts.append(
                self.collapsible(
                    self.H.div["snek-title-row", importance_class](
                        self.H.span(code.co_name),
                        self.H.span(
                            f"{format_libpath(filename)}:{fr.f_lineno}"
                        ),
                    ),
                    self(ed)
                    if ed is not None
                    else self.H.span("Could not find source"),
                )
            )
            curr = curr.tb_next

        return self.H.div["hrepr-body"](*parts, exc_proper)

    def hrepr(self, r: range):  # noqa: F811
        rval = self.H.instance(self(r.start), self(r.stop), type="range")
        if r.step != 1:
            rval = rval(self(r.step))
        return rval

    def hrepr(self, fn: types.BuiltinFunctionType):  # noqa: F811
        if self.state.depth == 0:
            return H.instance(
                H.pre["snek-docstring"](fn.__doc__),
                type=fn.__name__,
                vertical=True,
            )
        else:
            return NotImplemented

    def hrepr(self, fn: FunctionType):  # noqa: F811
        if self.state.depth == 0:
            ed = edit(fn)
            if ed is None:
                return NotImplemented
            return self(ed)
        else:
            return NotImplemented

    def hrepr(self, fn: MethodType):  # noqa: F811
        if self.state.depth == 0:
            ed = edit(fn.__func__)
            if ed is None:
                return NotImplemented
            return H.instance(
                self(ed), type=self.hrepr_short(fn), vertical=True
            )
        else:
            return NotImplemented

    def hrepr(self, mod: ModuleType):  # noqa: F811
        exclusions = {
            "__builtins__",
            "__name__",
            "__doc__",
            "__path__",
            "__file__",
            "__cached__",
            "__package__",
            "__loader__",
            "__spec__",
            "__all__",
            "__author__",
        }
        if self.state.depth == 0:
            return self.H.instance(
                self.H.pre(mod.__doc__.strip()) if mod.__doc__ else "",
                *[
                    self.H.pair(
                        represents(value, self.H.span(name)),
                        self(value, max_depth=2),
                        delimiter="=",
                    )
                    for name, value in sorted(vars(mod).items())
                    if name not in exclusions
                ],
                type=self.H.span(
                    self.H.span["snek-block-type"]("module "), mod.__name__
                ),
                vertical=True,
            )
        else:
            return NotImplemented

    def hrepr(self, cls: type):  # noqa: F811
        if self.state.depth > 0:
            return NotImplemented

        exclusions = {"__dict__", "__module__", "__weakref__", "__doc__"}

        # Exclude the object type to reduce noise
        mro = list(type.mro(cls))[:-1]

        rows = {}
        clselems = []
        for cls2 in mro:
            clsname = cls2.__qualname__
            clselem = represents(cls2, self.H.span(clsname))
            clselems.append(clselem)
            for name, value in vars(cls2).items():
                if name not in rows and name not in exclusions:
                    rows[name] = (
                        name,
                        clsname,
                        clselem,
                        represents(value, self.H.span(name)),
                        value,
                    )

        rows = list(rows.values())
        rows.sort(key=lambda row: row[0])

        tbl = self.H.table["hrepr-body"]()

        doc = getattr(cls, "__doc__", None)
        if doc:
            tbl = tbl(self.H.tr(self.H.td(self.H.pre(doc), colspan=3)))

        for _, clsname, clselem, nameelem, value in rows:
            css_class = (
                "snek-clsname-principal"
                if clsname == cls.__name__
                else "snek-clsname"
            )
            tbl = tbl(
                self.H.tr(
                    self.H.td[css_class](clselem, "."),
                    self.H.td(nameelem),
                    self.H.td("= ", self(value)),
                )
            )

        title = self.H.span(
            self.H.span["snek-block-type"]("class "), *join(clselems, " > ")
        )

        return self.H.instance(tbl, type=title, vertical=True)

    def hrepr_short(self, clsm: (classmethod, staticmethod)):  # noqa: F811
        return self.H.defn(type(clsm).__name__, shortname(clsm.__func__))

    def hrepr(self, clsm: (classmethod, staticmethod)):  # noqa: F811
        fn = clsm.__func__
        if self.state.depth == 0 and isinstance(fn, FunctionType):
            ed = edit(fn)
            if ed is None:
                return NotImplemented
            return self.H.div(self(ed))
        else:
            return NotImplemented

    def hrepr_short(self, prop: property):  # noqa: F811
        return self.H.defn(
            type(prop).__name__, shortname(prop.fget or prop.fset or prop.fdel)
        )

    def hrepr(self, prop: property):  # noqa: F811
        if self.state.depth == 0:
            title = "property"
            return self.H.instance(
                H.pair("fget", self(edit(prop.fget)), delimiter="="),
                H.pair("fset", self(edit(prop.fset)), delimiter="="),
                H.pair("fdel", self(edit(prop.fdel)), delimiter="="),
                type=title,
                vertical=True,
            )
        else:
            return NotImplemented


#####################################
# Inject changes into default hrepr #
#####################################


sktk = SimpleNamespace(
    edit=edit,
    help=None,
    imp=Importer(),
    mod=mod,
    fill_at=fill_at,
    insert_at=insert_at,
    clear_at=clear_at,
    pastevar=pastevar,
)


def snekbreakpoint():
    SnekTalkDb().set_trace()


@standard_html.variant(
    initial_state={"hjson": sktk_hjson, "requirejs_resources": []}
)
def sktk_html(self, node: type(H.include)):
    _, children, data = _extract_as(self, node, "include", path=None, type=None)
    if data.type is None or data.path is None:
        raise TypeError("H.include must have a type and a path")

    path = os.path.expanduser(data.path)
    path = os.path.abspath(path)

    if data.type == "text/css":
        return H.link(rel="stylesheet", href=f"/fs{path}")
    elif data.type == "text/javascript":
        return H.script(type="text/javascript", src=f"/fs{path}")
    else:
        raise TypeError(f"Cannot include type '{data.type}'")


def inject():
    builtins.help = help_placeholder
    builtins.print = snekprint_override
    builtins.breakpoint = snekbreakpoint
    builtins.hdir = hdir
    builtins.print0 = orig_print
    builtins.sktk = sktk
    hrepr.configure(
        mixins=SnekTalkHrepr, postprocess=wrap_onclick, backend=sktk_html,
    )



import importlib
import sys

from .debug import SnekTalkDb
from .evaluator import Evaluator
from .repr import inject, snekprint as print
from .server import serve
from .session import current_session
from .utils import Interactor, pastecode, pastevar
from .version import version


def interact(prompt=">>>", **kwargs):
    glb = sys._getframe(1).f_globals
    mname = glb["__name__"]
    module = importlib.import_module(mname)
    lcl = sys._getframe(1).f_locals
    if not (sess := current_session()):
        sess = serve(**kwargs, template={"title": mname})
    Evaluator(module, glb, lcl, sess, prompt=prompt).loop()


def debug(**kwargs):
    if not current_session():
        serve(**kwargs, template={"title": "debug"})
    SnekTalkDb().set_trace()
